//+------------------------------------------------------------------+
//|                                                   optimizers.mqh |
//|                                     Copyright 2023, Omega Joctan |
//|                        https://www.mql5.com/en/users/omegajoctan |
//+------------------------------------------------------------------+
#property copyright "Copyright 2023, Omega Joctan"
#property link      "https://www.mql5.com/en/users/omegajoctan"

//+------------------------------------------------------------------+
//|Class containing optimizers for updating neural network parameters|
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//|                                                                  |
//| Adam (Adaptive Moment Estimation): Adam is an adaptive learning  |
//| rate optimizer that maintains learning rates for each parameter  |
//| and adapts them based on the first and second moments of the     |
//| gradients. It's known for its fast convergence and robustness to |
//| noisy gradients                                                  |
//|                                                                  |
//+------------------------------------------------------------------+

/*enum Optimizer {
    OPTIMIZER_SGD, //Stochastic Gradient Descent
    OPTIMIZER_MiniBatchGD, //Mini-Batch Gradiend Descent
    OPTIMIZER_Adam, //Adaptive Moment Estimation
    OPTIMIZER_RMSprop, //Root Mean Square Propagation
    OPTIMIZER_Adagrad, //Adaptive Gradient Descent 
    OPTIMIZER_Adadelta, //Adadelta
    OPTIMIZER_Nadam //Nesterov-accelerated Adaptive Moment Estimation
};
*/

class OptimizerAdam
  {
protected:
   int time_step;
   double m_learning_rate;
   double m_beta1;
   double m_beta2;
   double m_epsilon;
   
   matrix m; //first moment estimate
   matrix v; //second moment estimate
   
public:
                     OptimizerAdam(double learning_rate=0.01, double beta1=0.9, double beta2=0.999, double epsilon=1e-8);
                    ~OptimizerAdam(void);
                    
                    virtual void update(matrix &parameters, matrix &gradients);
  };
//+------------------------------------------------------------------+
//|  Initializes the Adam optimizer with hyperparameters.            |   
//|                                                                  |
//|  learning_rate: Step size for parameter updates                  |
//|  beta1: Decay rate for the first moment estimate                 |
//|     (moving average of gradients).                               |
//|  beta2: Decay rate for the second moment estimate                |
//|     (moving average of squared gradients).                       |
//|  epsilon: Small value for numerical stability.                   |
//+------------------------------------------------------------------+
OptimizerAdam::OptimizerAdam(double learning_rate=0.010000, double beta1=0.9, double beta2=0.999, double epsilon=1e-8):
 time_step(0),
 m_learning_rate(learning_rate),
 m_beta1(beta1),
 m_beta2(beta2),
 m_epsilon(epsilon)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerAdam::~OptimizerAdam(void)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OptimizerAdam::update(matrix &parameters,matrix &gradients)
 {
    // Initialize m and v matrices if not already initialized
    if (m.Rows() != parameters.Rows() || m.Cols() != parameters.Cols())
    {
        m.Resize(parameters.Rows(), parameters.Cols());
        m.Fill(0.0);
    }

    if (v.Rows() != parameters.Rows() || v.Cols() != parameters.Cols())
    {
        v.Resize(parameters.Rows(), parameters.Cols());
        v.Fill(0.0);
    }

   
    this.time_step++; 
    
    this.m = this.m_beta1 * this.m + (1 -  this.m_beta1) * gradients;
    
    this.v = this.m_beta2 * this.v + (1 -  this.m_beta2) * MathPow(gradients, 2);

//--- Bias correction

    matrix m_hat = this.m / (1 - MathPow(this.m_beta1, this.time_step));
    
    matrix v_hat = this.v / (1 - MathPow(this.m_beta2, this.time_step));
    
    parameters -= (this.m_learning_rate * m_hat) / (MathPow(v_hat, 0.5) + this.m_epsilon);
 }
 
 
//+------------------------------------------------------------------+
//|                                                                  |
//|                                                                  |
//| Stochastic Gradient Descent (SGD):                               |
//|                                                                  |
//| This is the simplest optimizer                                   |
//| It updates the parameters using the gradients of the loss        |
//| function computed on individual training samples or mini-batches.|
//|                                                                  |
//|                                                                  |
//+------------------------------------------------------------------+


class OptimizerSGD
  {
protected:
   double m_learning_rate;
   
public:
                     OptimizerSGD(double learning_rate=0.01);
                    ~OptimizerSGD(void);
                     
                    virtual void update(matrix &parameters, matrix &gradients);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerSGD::OptimizerSGD(double learning_rate=0.01):
 m_learning_rate(learning_rate)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerSGD::~OptimizerSGD(void)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OptimizerSGD::update(matrix &parameters, matrix &gradients)
 {
    parameters -= this.m_learning_rate * gradients;
 }
 
 
//+------------------------------------------------------------------+
//|  Batch Gradient Descent (BGD): This optimizer computes the       |
//|  gradients of the loss function on the entire training dataset   |
//|  and updates the parameters accordingly. It can be slow and      |
//|  memory-intensive for large datasets but tends to provide a      |
//|  stable convergence.                                             |
//+------------------------------------------------------------------+


class OptimizerBGD: public OptimizerSGD
  {
public:
                     OptimizerBGD(double learning_rate=0.01);
                    ~OptimizerBGD(void);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerBGD::OptimizerBGD(double learning_rate=0.010000): OptimizerSGD(learning_rate)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerBGD::~OptimizerBGD(void)
 {
 
 }
 
//+------------------------------------------------------------------+
//|                                                                  |
//|                                                                  |
//|                                                                  |
//| RMSprop (Root Mean Square Propagation): RMSprop is similar to    |
//| Adam but only uses the first moment of the gradients to adapt the|
//| m learning rates. It's effective in handling non-stationary      |
//| objectives and can converge quickly for many problems.           |
//|                                                                  |
//|                                                                  |
//+------------------------------------------------------------------+


class OptimizerRMSprop
  {
protected:
   double m_learning_rate;
   double m_decay_rate;
   double m_epsilon;
   
   matrix<double> cache;
   
   //Dividing double/matrix causes compilation error | this is the fix to the issue
   matrix divide(const double numerator, const matrix &denominator)
    {
      matrix res = denominator;
      
      for (ulong i=0; i<denominator.Rows(); i++)
        res.Row(numerator / denominator.Row(i), i);
     return res;
    }
    
public:
                     OptimizerRMSprop(double learning_rate=0.01, double decay_rate=0.9, double epsilon=1e-8);
                    ~OptimizerRMSprop(void);
                    
                    virtual void update(matrix& parameters, matrix& gradients);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerRMSprop::OptimizerRMSprop(double learning_rate=0.01, double decay_rate=0.9, double epsilon=1e-8):
 m_learning_rate(learning_rate),
 m_decay_rate(decay_rate),
 m_epsilon(epsilon)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerRMSprop::~OptimizerRMSprop(void)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OptimizerRMSprop::update(matrix &parameters,matrix &gradients)
 {
 
   if (cache.Rows()!=parameters.Rows() || cache.Cols()!=parameters.Cols())
    {
     cache.Init(parameters.Rows(), parameters.Cols());
     cache.Fill(0.0);
    }
     
//---
    
    cache += m_decay_rate * cache + (1 - m_decay_rate) * MathPow(gradients, 2);
    parameters -= divide(m_learning_rate, cache + m_epsilon) * gradients;
 }


//+------------------------------------------------------------------+
//|                                                                  |
//|                                                                  |
//|                                                                  |
//| Adagrad (Adaptive Gradient Algorithm):                           |
//|                                                                  |
//| Adagrad adapts the learning rates for each parameter based on the|
//| historical gradients. It performs larger updates for infrequent  |
//| parameters and smaller updates for frequent parameters, making   |
//| it suitable for sparse data.                                     |
//|                                                                  |
//|                                                                  |
//|                                                                  |
//+------------------------------------------------------------------+

class OptimizerAdaGrad
  {
protected:
 double m_learning_rate;
 double m_epsilon;
 matrix cache;
 
 
   //Dividing double/matrix causes compilation error | this is the fix to the issue
   matrix divide(const double numerator, const matrix &denominator)
    {
      matrix res = denominator;
      
      for (ulong i=0; i<denominator.Rows(); i++)
        res.Row(numerator / denominator.Row(i), i);
     return res;
    }
 
public:
                     OptimizerAdaGrad(double learning_rate=0.01, double epsilon=1e-8);
                    ~OptimizerAdaGrad(void);
                    
                    virtual void update(matrix &parameters, matrix &gradients);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerAdaGrad::OptimizerAdaGrad(double learning_rate=0.01,double epsilon=1e-8):
 m_learning_rate(learning_rate),
 m_epsilon(epsilon)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerAdaGrad::~OptimizerAdaGrad(void)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OptimizerAdaGrad::update(matrix &parameters,matrix &gradients)
 {
   if (cache.Rows()!=parameters.Rows() || cache.Cols()!=parameters.Cols())
    {
     cache.Resize(parameters.Rows(), parameters.Cols());
     cache.Fill(0.0);
    }
     
//--- 
   
    cache += MathPow(gradients, 2);
    parameters -= divide(this.m_learning_rate,  cache + this.m_epsilon) * gradients;
 }

//+------------------------------------------------------------------+
//|                                                                  |
//| Adadelta:                                                        |      
//|                                                                  |
//| Adadelta is an extension of Adagrad that aims to address         |
//| its tendency to decrease the learning rate over time. It uses a  |
//| more sophisticated update rule that accounts for the accumulated |
//| gradients over a window of time.                                 |
//|                                                                  |
//+------------------------------------------------------------------+


class OptimizerAdaDelta
  {
protected:
   double m_rho;
   double m_epsilon;
   matrix E_g; //Exponential moving average of squared gradients
   matrix E_d; //Exponential moving average of squared parameter updates
   
public:
                     OptimizerAdaDelta(double rho=0.95, double epsilon=1e-8);
                    ~OptimizerAdaDelta(void);
                    
                    virtual void update(matrix &parameters, matrix &gradients);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerAdaDelta::OptimizerAdaDelta(double rho=0.95, double epsilon=1e-8):
  m_rho(rho),
  m_epsilon(epsilon)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerAdaDelta::~OptimizerAdaDelta(void)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OptimizerAdaDelta::update(matrix &parameters,matrix &gradients)
 {
    // Initialize m and v matrices if not already initialized
    if (E_g.Rows() != parameters.Rows() || E_g.Cols() != parameters.Cols())
    {
        E_g.Resize(parameters.Rows(), parameters.Cols());
        E_g.Fill(0.0);
    }

    if (E_d.Rows() != parameters.Rows() || E_d.Cols() != parameters.Cols())
    {
        E_d.Resize(parameters.Rows(), parameters.Cols());
        E_d.Fill(0.0);
    }
//---
   
   E_g = m_rho * this.E_g + (1 - m_rho) * MathPow(gradients, 2);
   matrix delta = -1 * (MathPow(E_d + this.m_epsilon, 0.5) / MathPow(E_g + this.m_epsilon, 0.5) ) * gradients;
   parameters += delta;
   
   this.E_d = m_rho * this.E_d + (1 - m_rho) * MathPow(delta, 2);
 }

//+------------------------------------------------------------------+
//|                                                                  |
//|                                                                  |
//| Nadam (Nesterov-accelerated Adaptive Moment Estimation):         |
//|                                                                  |
//| Nadam combines Nesterov momentum with Adam optimizer, which      |
//| results in faster convergence and better generalization.         |
//|                                                                  |
//|                                                                  |
//+------------------------------------------------------------------+



class OptimizerNadam
  {
protected:
   int time_step;
   double m_learning_rate;
   double m_beta1;
   double m_beta2;
   double m_epsilon;
   
   matrix m; //first moment estimate
   matrix v; //second moment estimate
   
   //Dividing double/matrix causes compilation error | this is the fix to the issue
   matrix divide(const double numerator, const matrix &denominator)
    {
      matrix res = denominator;
      
      for (ulong i=0; i<denominator.Rows(); i++)
        res.Row(numerator / denominator.Row(i), i);
     return res;
    }
    
public:
                     OptimizerNadam(double learning_rate=0.01, double beta1=0.9, double beta2=0.999, double epsilon=1e-8);
                    ~OptimizerNadam(void);
                    
                    virtual void update(matrix &parameters, matrix &gradients);
  };
//+------------------------------------------------------------------+
//|  Initializes the Adam optimizer with hyperparameters.            |   
//|                                                                  |
//|  learning_rate: Step size for parameter updates                  |
//|  beta1: Decay rate for the first moment estimate                 |
//|     (moving average of gradients).                               |
//|  beta2: Decay rate for the second moment estimate                |
//|     (moving average of squared gradients).                       |
//|  epsilon: Small value for numerical stability.                   |
//+------------------------------------------------------------------+
OptimizerNadam::OptimizerNadam(double learning_rate=0.010000, double beta1=0.9, double beta2=0.999, double epsilon=1e-8):
 time_step(0),
 m_learning_rate(learning_rate),
 m_beta1(beta1),
 m_beta2(beta2),
 m_epsilon(epsilon)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OptimizerNadam::~OptimizerNadam(void)
 {
 
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OptimizerNadam::update(matrix &parameters,matrix &gradients)
 {
    // Initialize m and v matrices if not already initialized
    if (m.Rows() != parameters.Rows() || m.Cols() != parameters.Cols())
    {
        m.Resize(parameters.Rows(), parameters.Cols());
        m.Fill(0.0);
    }

    if (v.Rows() != parameters.Rows() || v.Cols() != parameters.Cols())
    {
        v.Resize(parameters.Rows(), parameters.Cols());
        v.Fill(0.0);
    }

   
    this.time_step++; 
    
    this.m = this.m_beta1 * this.m + (1 -  this.m_beta1) * gradients;
    
    this.v = this.m_beta2 * this.v + (1 -  this.m_beta2) * MathPow(gradients, 2);
    
//--- Bias correction

    matrix m_hat = this.m / (1 - MathPow(this.m_beta1, this.time_step));
    matrix v_hat = this.v / (1 - MathPow(this.m_beta2, this.time_step));
    
    parameters -= divide(this.m_learning_rate , (MathPow(v_hat, 0.5) + this.m_epsilon)) * (this.m_beta1 * m_hat + (1 - this.m_beta1) * gradients / (1 - MathPow(this.m_beta1, time_step)));
 }

 
