//+------------------------------------------------------------------+
//|                                                 Pattern Nets.mqh |
//|                                    Copyright 2022, Fxalgebra.com |
//|                        https://www.mql5.com/en/users/omegajoctan |
//+------------------------------------------------------------------+
#property copyright "Copyright 2022, Fxalgebra.com"
#property link      "https://www.mql5.com/en/users/omegajoctan"
//+------------------------------------------------------------------+
//| Neural network type for pattern recognition/ can be used to      |
//| to predict discrete data target variables. They are widely known |
//| as classification Neural Networks                                |
//+------------------------------------------------------------------+
#include <MatrixExtend.mqh>
#include <metrics.mqh>
#include <Tensors.mqh>
#include <cross_validation.mqh>
#include <plots.mqh>
#include "optimizers.mqh"
#include "initializers.mqh"

#ifndef RANDOM_STATE
#define  RANDOM_STATE 42
#endif

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum activation
  {
   AF_HARD_SIGMOID_ = AF_HARD_SIGMOID,
   AF_SIGMOID_ = AF_SIGMOID,
   AF_SWISH_ = AF_SWISH,
   AF_SOFTSIGN_ = AF_SOFTSIGN,
   AF_TANH_ = AF_TANH,
   AF_ELU_ = AF_ELU,
   AF_EXP_ = AF_EXP,
   AF_GELU_ = AF_GELU,
   AF_LINEAR_ = AF_LINEAR,
   AF_LRELU_ = AF_LRELU,
   AF_RELU_ = AF_RELU,
   AF_SELU_ = AF_SELU,
   AF_TRELU_ = AF_TRELU,
   AF_SOFTPLUS_ = AF_SOFTPLUS
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+

enum loss
  {
    LOSS_BCE_ = LOSS_BCE, //Binary cross entropy
    LOSS_CCE_ = LOSS_CCE, //Categorical cross entropy
    LOSS_MSE_ = LOSS_MSE, //Mean squared error
    LOSS_HINGE_ = LOSS_HINGE, //Hinge
    LOSS_CAT_HINGE_ = LOSS_CAT_HINGE, //Categorical Hinge
  };

struct backprop //This structure returns the loss information obtained from the backpropagation function
  {
    vector training_loss,
           validation_loss;
           
           void Init(ulong epochs)
            {
              training_loss.Resize(epochs);
              validation_loss.Resize(epochs);
            }
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CPatternNets
  {
private:   
   CTensors          *Weights_tensor; //Weight Get
   CTensors          *Bias_tensor; //Bias Get
   CTensors          *Input_tensor;
   CTensors          *Output_tensor;
   
//---

   struct nn_config {
      ulong hidden_layers, inputs, outputs;
      vector classes_in_data;
      activation activation_fx;
      loss   loss_fx;
      vector w_config;
      vector HL_NODES;
   } config;
   
//---
   
protected:
   bool              isBackProp;
      
   string  ConvertTime(double seconds);
   matrix<double>    Partial_Derivatives;  
   bool loadedModel;
   bool trained;
   
   backprop          backpropagation(const matrix &x, const vector &y, const uint epochs, OptimizerAdam *adam, uint batch_size=0, bool show_batch_progress=false,weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER);
   backprop          backpropagation(const matrix &x, const vector &y, const uint epochs, OptimizerAdaDelta *adadelta, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER);
   backprop          backpropagation(const matrix &x, const vector &y, const uint epochs, OptimizerAdaGrad *adagrad, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER);
   backprop          backpropagation(const matrix &x, const vector &y, const uint epochs, OptimizerNadam *nadam, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER);
   backprop          backpropagation(const matrix &x, const vector &y, const uint epochs, OptimizerRMSprop *rmsprop, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER);
   backprop          backpropagation(const matrix &x, const vector &y, const uint epochs, OptimizerBGD *bgd, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER);
   backprop          backpropagation(const matrix &x, const vector &y, const uint epochs, OptimizerSGD *sgd, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER);
   
public:
                     CPatternNets(vector &HL_NODES, activation AF_, loss LOSS_); //For Training the model
                     CPatternNets(const double &architecture[], const double &activation_[], const double &classes_[], const double &inputs[], const double &weights[], const double &bias[]); //for loading the pre-trained model | suitable for production, models will be loaded from resources arrays extracted from binary files
                     
                    ~CPatternNets(void);
                    
   vector            predict_proba(const vector &x);
   matrix            predict_proba(const matrix &x);
   
   int               predict_bin(const vector &x);
   vector            predict_bin(const matrix &x);
   
   void              fit(const matrix &x, const vector &y, const uint epochs, OptimizerAdam *adam, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER); //Backpropagation with adam
   void              fit(const matrix &x, const vector &y, const uint epochs, OptimizerAdaDelta *adadelta, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER); //Backpropagation with adadelta
   void              fit(const matrix &x, const vector &y, const uint epochs, OptimizerAdaGrad *adagrad, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER); //Backpropagation with adagrad
   void              fit(const matrix &x, const vector &y, const uint epochs, OptimizerNadam *nadam, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER); //Backpropagation with nadam
   void              fit(const matrix &x, const vector &y, const uint epochs, OptimizerRMSprop *rmsprop, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER); //Backpropagation with rmsprop
   void              fit(const matrix &x, const vector &y, const uint epochs, OptimizerBGD *bgd, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER); //Backpropagation with BGD
   void              fit(const matrix &x, const vector &y, const uint epochs, OptimizerSGD *sgd, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER); //Backpropagation with SGD
   
   bool              save_model(string dir);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CPatternNets::CPatternNets(vector &HL_NODES, activation AF_, loss LOSS_)
  {
//--- Pattern nets configurations
   
   this.config.activation_fx = AF_;
   this.config.loss_fx = LOSS_;
   this.config.hidden_layers = HL_NODES.Size();
   this.config.HL_NODES = HL_NODES;
   trained = false;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CPatternNets::~CPatternNets(void)
  {
     if (CheckPointer(Weights_tensor)!=POINTER_INVALID) delete (Weights_tensor);
     
     if (CheckPointer(Bias_tensor)!=POINTER_INVALID) delete (Bias_tensor);
     
     if (CheckPointer(Input_tensor)!=POINTER_INVALID) delete (Input_tensor);
     
     if (CheckPointer(Output_tensor)!=POINTER_INVALID) delete (Output_tensor);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
vector CPatternNets::predict_proba(const vector &x)
 {
  
  if (!trained)
   {
     printf("%s Train the model first before using it to make predictions | call the fit function first",__FUNCTION__);
     vector empty = {};
     return empty;
   }
  
  
  if (x.Size() != config.inputs)
   {
      printf("%s Dimensions mistmatch | Betweeen the new data and the data used during training",__FUNCTION__);
      
      vector empty = {};
      return empty;
   }
  
  matrix L_INPUT = MatrixExtend::VectorToMatrix(x);
  matrix L_OUTPUT = {};
  
  for (ulong i = 0; i < config.hidden_layers; i++)
    {
       if (isBackProp) this.Input_tensor.Add(L_INPUT, i);
   
       L_OUTPUT = this.Weights_tensor.Get(i).MatMul(L_INPUT) + this.Bias_tensor.Get(i); //Weights x Inputs + Bias
       
       L_OUTPUT.Activation(L_OUTPUT, i+1==config.hidden_layers?AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx)); // Activation
       
       L_INPUT = L_OUTPUT; //Next layer input is previous layer outputs
   
       if (isBackProp) this.Output_tensor.Add(L_OUTPUT, i);
   }
//---
  
  return MatrixExtend::MatrixToVector(L_OUTPUT);
}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int CPatternNets::predict_bin(const vector &x)
 {
   vector ret_v = this.predict_proba(x);
   
   return ((int)config.classes_in_data[ret_v.ArgMax()]); //Assign a class to the highest probability
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
matrix CPatternNets::predict_proba(const matrix &x)
  {
   matrix mat(x.Rows(), config.outputs);
   vector v(config.outputs);
   
   for(ulong i=0; i<x.Rows(); i++)
      {
         v = this.predict_proba(x.Row(i));
         mat.Row(v, i);
      }
   return (mat);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
vector CPatternNets::predict_bin(const matrix &x)
 {
   ulong rws = x.Rows();
   
   vector ret_v(rws);
   for (ulong i=0; i<rws; i++)
      ret_v[i] = this.predict_bin(x.Row(i));
   
   return (ret_v);
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+

string CPatternNets::ConvertTime(double seconds)
{
    string time_str = "";
    uint minutes = 0, hours = 0;

    if (seconds >= 60)
    {
        minutes = (uint)(seconds / 60.0) ;
        seconds = fmod(seconds, 1.0) * 60;
        time_str = StringFormat("%d Minutes and %.3f Seconds", minutes, seconds);
    }
    
    if (minutes >= 60)
    {
        hours = (uint)(minutes / 60.0);
        minutes = minutes % 60;
        time_str = StringFormat("%d Hours and %d Minutes", hours, minutes);
    }

    if (time_str == "")
    {
        time_str = StringFormat("%.3f Seconds", seconds);
    }

    return time_str;
}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
backprop CPatternNets::backpropagation(const matrix &x,const vector &y,const uint epochs,OptimizerAdaDelta *adadelta,uint batch_size=0,bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {
   
//--- dataframe configurations
   
   ulong inputs = x.Cols();
   ulong rows  = x.Rows();

   config.classes_in_data = MatrixExtend::Unique(y);
   config.outputs = (int)config.classes_in_data.Size();
   config.inputs = inputs;
   
//---

   isBackProp = false;
   
   backprop backprop_struct;
   backprop_struct.Init(epochs);
   
   if (y.Size() != rows )
     {
        Print(__FUNCTION__," FATAL | Number of rows  in the x matrix is not the same the y vector size ");
        
        return backprop_struct;
     }
     
     vector v2 = {(int)config.outputs};
     config.HL_NODES = MatrixExtend::concatenate(config.HL_NODES, v2); //Appending the output layer size to the hidden layers 

     config.w_config.Resize(config.HL_NODES.Size()); 
     config.hidden_layers = config.HL_NODES.Size();

//--- Initiating the Optimizer
   
    OptimizerAdaDelta optimizer_weights = adadelta;
    OptimizerAdaDelta optimizer_bias = adadelta;

//--- Generating the weights and biases
    
     this.Weights_tensor = new CTensors((uint)config.hidden_layers);
     this.Bias_tensor = new CTensors((uint)config.hidden_layers);
     this.Input_tensor = new CTensors((uint)config.hidden_layers);
     this.Output_tensor = new CTensors((uint)config.hidden_layers);
     
     ulong layer_input = config.inputs; 
     
     matrix W, B;
     
     for (ulong i=0; i<config.hidden_layers; i++)
       {
          config.w_config[i] = layer_input*config.HL_NODES[i];
          
          W = WeightsInitializers::Initializer((ulong)config.HL_NODES[i], layer_input, WEIGHT_INIT);
          this.Weights_tensor.Add(W, i);
          
          B.Init((ulong)config.HL_NODES[i], 1);
          B.Fill(0.0); //Random weight initialization can lead to symmetry issues where all neurons in a layer learn identical features. Zeros for biases help break this symmetry, allowing each neuron to learn independently during training.
          this.Bias_tensor.Add(B, i);
          
          layer_input = (ulong)config.HL_NODES[i];
       }
   
   isBackProp = true;
    
//---      
   
    matrix DELTA = {};
    vector Forecast_v={};
    
    matrix  temp_inputs ={};
   
    
    matrix dB = {}; //Bias Derivatives
    matrix dW = {}; //Weight Derivatives 
    
    
    matrix ONEHOT = MatrixExtend::OneHotEncoding(y); //ONE HOT ENCODED MATRIX
    
//---
    
    CCrossValidation cross_validation;        
    CTensors *cv_tensor;
    matrix validation_data = MatrixExtend::concatenate(x, y);
    matrix validation_x;
    vector validation_y;
     
    cv_tensor =  cross_validation.KFoldCV(validation_data, 10);
    
    for (ulong epoch=0; epoch<epochs && !IsStopped(); epoch++) //Run an entire epoch
      { 
       
        double epoch_start = GetTickCount(); 
        
        uint num_batches = (uint)MathFloor(x.Rows()/(batch_size+DBL_EPSILON));
        
        vector batch_loss(num_batches), 
               batch_accuracy(num_batches);
               
        vector actual_v ={}, pred_v={}, LossGradient = {};
        
        if (batch_size>0)
         {  
            for (uint batch=0, batch_start=0, batch_end=batch_size; batch<num_batches; batch++, batch_start+=batch_size, batch_end=(batch_start+batch_size-1))
               {
                  
                  matrix batch_x = MatrixExtend::Get(x, batch_start, batch_end-1);
                  vector batch_y = MatrixExtend::Get(y, batch_start, batch_end-1);
                  
                  rows = batch_x.Rows();            
                  matrix ONEHOT = MatrixExtend::OneHotEncoding(batch_y); //ONE HOT ENCODED MATRIX      
                  
                    for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                      {
                        pred_v = predict_proba(batch_x.Row(iter));
                        actual_v = ONEHOT.Row(iter);
                        
            //---
                        
                         DELTA.Resize(config.outputs, 1);
                         
                         for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                           {                
                              Partial_Derivatives = this.Output_tensor.Get(int(layer));
                              temp_inputs = this.Input_tensor.Get(int(layer));
                              
                              Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                              
                              if (config.hidden_layers-1 == layer) //Last layer
                               {                     
                                 LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                                  
                                 DELTA.Col(LossGradient, 0);
                               }
                               
                              else
                               {
                                 W = this.Weights_tensor.Get(layer+1);
                                 
                                 DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                               }
                         
                           W = this.Weights_tensor.Get(layer);
                           B = this.Bias_tensor.Get(layer);
                           
                           dB = DELTA;
                           dW = DELTA.MatMul(temp_inputs.Transpose());  
                            
                         //--- Updating the weights using the chosen algorithm
                           
                           optimizer_weights.update(W, dW);
                           optimizer_bias.update(B, dB);
                           
                           this.Weights_tensor.Add(W, layer);
                           this.Bias_tensor.Add(B, layer);  
                        }
                   }
                 
                 pred_v = predict_bin(batch_x);
                 
                 batch_loss[batch] = pred_v.Loss(batch_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
                 batch_loss[batch] = MathIsValidNumber(batch_loss[batch]) ? (batch_loss[batch]>1e6 ? 1e6 : batch_loss[batch]) : 1e6; //Check for nan and return some large value if it is nan
                 
                 batch_accuracy[batch] = Metrics::accuracy_score(batch_y,pred_v);
                 
                 if (show_batch_progress)
                  printf("----> batch[%d/%d] batch-loss %.5f accuracy %.3f",batch+1,num_batches,batch_loss[batch],batch_accuracy[batch]);  
              }
           }    
         else
           {
              for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                {
                  pred_v = predict_proba(x.Row(iter));
                  actual_v = ONEHOT.Row(iter);
                  
      //---
                   DELTA.Resize(config.outputs, 1);
                   
                   for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                     {                
                        Partial_Derivatives = this.Output_tensor.Get(int(layer));
                        temp_inputs = this.Input_tensor.Get(int(layer));
                        
                        Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                        
                        if (config.hidden_layers-1 == layer) //Last layer
                         {                     
                           LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                            
                           DELTA.Col(LossGradient, 0);
                         }
                         
                        else
                         {
                           W = this.Weights_tensor.Get(layer+1);
                           
                           DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                         }
                   
                        W = this.Weights_tensor.Get(layer);
                        B = this.Bias_tensor.Get(layer);
                        
                        dB = DELTA;
                        dW = DELTA.MatMul(temp_inputs.Transpose());  
                         
                      //--- Updating the weights using the chosen algorithm
                        
                        optimizer_weights.update(W, dW);
                        optimizer_bias.update(B, dB);
                        
                        this.Weights_tensor.Add(W, layer);
                        this.Bias_tensor.Add(B, layer);  
                     
                    }
                 }  
             }
             
      //--- End of an epoch
      
        vector validation_loss(cv_tensor.SIZE);
        vector validation_acc(cv_tensor.SIZE);
        for (ulong i=0; i<cv_tensor.SIZE; i++)
          {
            validation_data = cv_tensor.Get(i);
            MatrixExtend::XandYSplitMatrices(validation_data, validation_x, validation_y);
            
            vector val_preds = this.predict_bin(validation_x);;
            
            validation_loss[i] = val_preds.Loss(validation_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
            validation_acc[i] = Metrics::accuracy_score(validation_y, val_preds);
          }
        
        
//---

        pred_v = this.predict_bin(x);
        
        if (batch_size==0)
          {              
              backprop_struct.training_loss[epoch] = pred_v.Loss(y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
        else
          {
              backprop_struct.training_loss[epoch] = batch_loss.Mean();
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
          
        double epoch_stop = GetTickCount();  
        
        printf("--> Epoch [%d/%d] training -> loss %.8f accuracy_score %.3f validation -> loss %.5f accuracy_score %.3f | Elapsed %s ",epoch+1,epochs,backprop_struct.training_loss[epoch],Metrics::accuracy_score(y, pred_v),backprop_struct.validation_loss[epoch],validation_acc.Mean(),this.ConvertTime((epoch_stop-epoch_start)/1000.0)); 
    }  
  
  
  if (CheckPointer(adadelta)!=POINTER_INVALID)  
    delete adadelta;
      
  return backprop_struct;
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
backprop CPatternNets::backpropagation(const matrix &x,const vector &y,const uint epochs,OptimizerAdaGrad *adagrad,uint batch_size=0,bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {
   
//--- dataframe configurations
   
   ulong inputs = x.Cols();
   ulong rows  = x.Rows();

   config.classes_in_data = MatrixExtend::Unique(y);
   config.outputs = (int)config.classes_in_data.Size();
   config.inputs = inputs;
   
//---

   isBackProp = false;
   
   backprop backprop_struct;
   backprop_struct.Init(epochs);
   
   if (y.Size() != rows )
     {
        Print(__FUNCTION__," FATAL | Number of rows  in the x matrix is not the same the y vector size ");
        
        return backprop_struct;
     }
     
     vector v2 = {(int)config.outputs};
     config.HL_NODES = MatrixExtend::concatenate(config.HL_NODES, v2); //Appending the output layer size to the hidden layers 

     config.w_config.Resize(config.HL_NODES.Size()); 
     config.hidden_layers = config.HL_NODES.Size();

//--- Initiating the Optimizer
   
    OptimizerAdaGrad optimizer_weights = adagrad;
    OptimizerAdaGrad optimizer_bias = adagrad;

//--- Generating the weights and biases
    
     this.Weights_tensor = new CTensors((uint)config.hidden_layers);
     this.Bias_tensor = new CTensors((uint)config.hidden_layers);
     this.Input_tensor = new CTensors((uint)config.hidden_layers);
     this.Output_tensor = new CTensors((uint)config.hidden_layers);
     
     ulong layer_input = config.inputs; 
     
     matrix W, B;
     
     for (ulong i=0; i<config.hidden_layers; i++)
       {
          config.w_config[i] = layer_input*config.HL_NODES[i];
          
          W = WeightsInitializers::Initializer((ulong)config.HL_NODES[i], layer_input, WEIGHT_INIT);
          this.Weights_tensor.Add(W, i);
          
          B.Init((ulong)config.HL_NODES[i], 1);
          B.Fill(0.0); //Random weight initialization can lead to symmetry issues where all neurons in a layer learn identical features. Zeros for biases help break this symmetry, allowing each neuron to learn independently during training.
          this.Bias_tensor.Add(B, i);
          
          layer_input = (ulong)config.HL_NODES[i];
       }
   
   isBackProp = true;
    
//---      
   
    matrix DELTA = {};
    vector Forecast_v={};
    
    matrix  temp_inputs ={};
   
    
    matrix dB = {}; //Bias Derivatives
    matrix dW = {}; //Weight Derivatives 
    
    
    matrix ONEHOT = MatrixExtend::OneHotEncoding(y); //ONE HOT ENCODED MATRIX
    
//---
    
    CCrossValidation cross_validation;        
    CTensors *cv_tensor;
    matrix validation_data = MatrixExtend::concatenate(x, y);
    matrix validation_x;
    vector validation_y;
     
    cv_tensor =  cross_validation.KFoldCV(validation_data, 10);
    
    for (ulong epoch=0; epoch<epochs && !IsStopped(); epoch++) //Run an entire epoch
      { 
       
        double epoch_start = GetTickCount(); 
        
        uint num_batches = (uint)MathFloor(x.Rows()/(batch_size+DBL_EPSILON));
        
        vector batch_loss(num_batches), 
               batch_accuracy(num_batches);
               
        vector actual_v ={}, pred_v={}, LossGradient = {};
        
        if (batch_size>0)
         {  
            for (uint batch=0, batch_start=0, batch_end=batch_size; batch<num_batches; batch++, batch_start+=batch_size, batch_end=(batch_start+batch_size-1))
               {
                  
                  matrix batch_x = MatrixExtend::Get(x, batch_start, batch_end-1);
                  vector batch_y = MatrixExtend::Get(y, batch_start, batch_end-1);
                  
                  rows = batch_x.Rows();            
                  matrix ONEHOT = MatrixExtend::OneHotEncoding(batch_y); //ONE HOT ENCODED MATRIX      
                  
                    for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                      {
                        pred_v = predict_proba(batch_x.Row(iter));
                        actual_v = ONEHOT.Row(iter);
                        
            //---
                        
                         DELTA.Resize(config.outputs, 1);
                         
                         for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                           {                
                              Partial_Derivatives = this.Output_tensor.Get(int(layer));
                              temp_inputs = this.Input_tensor.Get(int(layer));
                              
                              Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                              
                              if (config.hidden_layers-1 == layer) //Last layer
                               {                     
                                 LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                                  
                                 DELTA.Col(LossGradient, 0);
                               }
                               
                              else
                               {
                                 W = this.Weights_tensor.Get(layer+1);
                                 
                                 DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                               }
                         
                           W = this.Weights_tensor.Get(layer);
                           B = this.Bias_tensor.Get(layer);
                           
                           dB = DELTA;
                           dW = DELTA.MatMul(temp_inputs.Transpose());  
                            
                         //--- Updating the weights using the chosen algorithm
                           
                           optimizer_weights.update(W, dW);
                           optimizer_bias.update(B, dB);
                           
                           this.Weights_tensor.Add(W, layer);
                           this.Bias_tensor.Add(B, layer);  
                        }
                   }
                 
                 pred_v = predict_bin(batch_x);
                 
                 batch_loss[batch] = pred_v.Loss(batch_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
                 batch_loss[batch] = MathIsValidNumber(batch_loss[batch]) ? (batch_loss[batch]>1e6 ? 1e6 : batch_loss[batch]) : 1e6; //Check for nan and return some large value if it is nan
                 
                 batch_accuracy[batch] = Metrics::accuracy_score(batch_y,pred_v);
                 
                 if (show_batch_progress)
                  printf("----> batch[%d/%d] batch-loss %.5f accuracy %.3f",batch+1,num_batches,batch_loss[batch],batch_accuracy[batch]);  
              }
           }    
         else
           {
              for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                {
                  pred_v = predict_proba(x.Row(iter));
                  actual_v = ONEHOT.Row(iter);
                  
      //---
                   DELTA.Resize(config.outputs, 1);
                   
                   for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                     {                
                        Partial_Derivatives = this.Output_tensor.Get(int(layer));
                        temp_inputs = this.Input_tensor.Get(int(layer));
                        
                        Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                        
                        if (config.hidden_layers-1 == layer) //Last layer
                         {                     
                           LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                            
                           DELTA.Col(LossGradient, 0);
                         }
                         
                        else
                         {
                           W = this.Weights_tensor.Get(layer+1);
                           
                           DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                         }
                   
                        W = this.Weights_tensor.Get(layer);
                        B = this.Bias_tensor.Get(layer);
                        
                        dB = DELTA;
                        dW = DELTA.MatMul(temp_inputs.Transpose());  
                         
                      //--- Updating the weights using the chosen algorithm
                        
                        optimizer_weights.update(W, dW);
                        optimizer_bias.update(B, dB);
                        
                        this.Weights_tensor.Add(W, layer);
                        this.Bias_tensor.Add(B, layer);  
                     
                    }
                 }  
             }
             
      //--- End of an epoch
      
        vector validation_loss(cv_tensor.SIZE);
        vector validation_acc(cv_tensor.SIZE);
        for (ulong i=0; i<cv_tensor.SIZE; i++)
          {
            validation_data = cv_tensor.Get(i);
            MatrixExtend::XandYSplitMatrices(validation_data, validation_x, validation_y);
            
            vector val_preds = this.predict_bin(validation_x);;
            
            validation_loss[i] = val_preds.Loss(validation_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
            validation_acc[i] = Metrics::accuracy_score(validation_y, val_preds);
          }
      
//---

        pred_v = this.predict_bin(x);
        
        if (batch_size==0)
          {              
              backprop_struct.training_loss[epoch] = pred_v.Loss(y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
        else
          {
              backprop_struct.training_loss[epoch] = batch_loss.Mean();
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
          
        double epoch_stop = GetTickCount();  
        
        printf("--> Epoch [%d/%d] training -> loss %.8f accuracy_score %.3f validation -> loss %.5f accuracy_score %.3f | Elapsed %s ",epoch+1,epochs,backprop_struct.training_loss[epoch],Metrics::accuracy_score(y, pred_v),backprop_struct.validation_loss[epoch],validation_acc.Mean(),this.ConvertTime((epoch_stop-epoch_start)/1000.0)); 
    }  
  
  
  if (CheckPointer(adagrad)!=POINTER_INVALID)  
    delete adagrad;
      
  return backprop_struct;
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
backprop CPatternNets::backpropagation(const matrix &x,const vector &y,const uint epochs,OptimizerAdam *adam,uint batch_size=0,bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {
   
//--- dataframe configurations
   
   ulong inputs = x.Cols();
   ulong rows  = x.Rows();

   config.classes_in_data = MatrixExtend::Unique(y);
   config.outputs = (int)config.classes_in_data.Size();
   config.inputs = inputs;
   
//---

   isBackProp = false;
   
   backprop backprop_struct;
   backprop_struct.Init(epochs);
   
   if (y.Size() != rows )
     {
        Print(__FUNCTION__," FATAL | Number of rows  in the x matrix is not the same the y vector size ");
        
        return backprop_struct;
     }
     
     vector v2 = {(int)config.outputs};
     config.HL_NODES = MatrixExtend::concatenate(config.HL_NODES, v2); //Appending the output layer size to the hidden layers 

     config.w_config.Resize(config.HL_NODES.Size()); 
     config.hidden_layers = config.HL_NODES.Size();

//--- Initiating the Optimizer
   
    OptimizerAdam optimizer_weights = adam;
    OptimizerAdam optimizer_bias = adam;

//--- Generating the weights and biases
    
     this.Weights_tensor = new CTensors((uint)config.hidden_layers);
     this.Bias_tensor = new CTensors((uint)config.hidden_layers);
     this.Input_tensor = new CTensors((uint)config.hidden_layers);
     this.Output_tensor = new CTensors((uint)config.hidden_layers);
     
     ulong layer_input = config.inputs; 
     
     matrix W, B;
     
     for (ulong i=0; i<config.hidden_layers; i++)
       {
          config.w_config[i] = layer_input*config.HL_NODES[i];
          
          W = WeightsInitializers::Initializer((ulong)config.HL_NODES[i], layer_input, WEIGHT_INIT);
          this.Weights_tensor.Add(W, i);
          
          B.Init((ulong)config.HL_NODES[i], 1);
          B.Fill(0.0); //Random weight initialization can lead to symmetry issues where all neurons in a layer learn identical features. Zeros for biases help break this symmetry, allowing each neuron to learn independently during training.
          this.Bias_tensor.Add(B, i);
          
          layer_input = (ulong)config.HL_NODES[i];
       }
   
   isBackProp = true;
    
//---      
   
    matrix DELTA = {};
    vector Forecast_v={};
    
    matrix  temp_inputs ={};
   
    
    matrix dB = {}; //Bias Derivatives
    matrix dW = {}; //Weight Derivatives 
    
    
    matrix ONEHOT = MatrixExtend::OneHotEncoding(y); //ONE HOT ENCODED MATRIX
    
//---
    
    CCrossValidation cross_validation;        
    CTensors *cv_tensor;
    matrix validation_data = MatrixExtend::concatenate(x, y);
    matrix validation_x;
    vector validation_y;
     
    cv_tensor =  cross_validation.KFoldCV(validation_data, 10);
    
    for (ulong epoch=0; epoch<epochs && !IsStopped(); epoch++) //Run an entire epoch
      { 
       
        double epoch_start = GetTickCount(); 
        
        uint num_batches = (uint)MathFloor(x.Rows()/(batch_size+DBL_EPSILON));
        
        vector batch_loss(num_batches), 
               batch_accuracy(num_batches);
               
        vector actual_v ={}, pred_v={}, LossGradient = {};
        
        if (batch_size>0)
         {  
            for (uint batch=0, batch_start=0, batch_end=batch_size; batch<num_batches; batch++, batch_start+=batch_size, batch_end=(batch_start+batch_size-1))
               {
                  
                  matrix batch_x = MatrixExtend::Get(x, batch_start, batch_end-1);
                  vector batch_y = MatrixExtend::Get(y, batch_start, batch_end-1);
                  
                  rows = batch_x.Rows();            
                  matrix ONEHOT = MatrixExtend::OneHotEncoding(batch_y); //ONE HOT ENCODED MATRIX      
                  
                    for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                      {
                        pred_v = predict_proba(batch_x.Row(iter));
                        actual_v = ONEHOT.Row(iter);
                        
            //---
                        
                         DELTA.Resize(config.outputs, 1);
                         
                         for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                           {                
                              Partial_Derivatives = this.Output_tensor.Get(int(layer));
                              temp_inputs = this.Input_tensor.Get(int(layer));
                              
                              Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                              
                              if (config.hidden_layers-1 == layer) //Last layer
                               {                     
                                 LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                                  
                                 DELTA.Col(LossGradient, 0);
                               }
                               
                              else
                               {
                                 W = this.Weights_tensor.Get(layer+1);
                                 
                                 DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                               }
                         
                           W = this.Weights_tensor.Get(layer);
                           B = this.Bias_tensor.Get(layer);
                           
                           dB = DELTA;
                           dW = DELTA.MatMul(temp_inputs.Transpose());  
                            
                         //--- Updating the weights using the chosen algorithm
                           
                           optimizer_weights.update(W, dW);
                           optimizer_bias.update(B, dB);
                           
                           this.Weights_tensor.Add(W, layer);
                           this.Bias_tensor.Add(B, layer);  
                        }
                   }
                 
                 pred_v = predict_bin(batch_x);
                 
                 batch_loss[batch] = pred_v.Loss(batch_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
                 batch_loss[batch] = MathIsValidNumber(batch_loss[batch]) ? (batch_loss[batch]>1e6 ? 1e6 : batch_loss[batch]) : 1e6; //Check for nan and return some large value if it is nan
                 
                 batch_accuracy[batch] = Metrics::accuracy_score(batch_y,pred_v);
                 
                 if (show_batch_progress)
                  printf("----> batch[%d/%d] batch-loss %.5f accuracy %.3f",batch+1,num_batches,batch_loss[batch],batch_accuracy[batch]);  
              }
           }    
         else
           {
              for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                {
                  pred_v = predict_proba(x.Row(iter));
                  actual_v = ONEHOT.Row(iter);
                  
      //---
                   DELTA.Resize(config.outputs, 1);
                   
                   for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                     {                
                        Partial_Derivatives = this.Output_tensor.Get(int(layer));
                        temp_inputs = this.Input_tensor.Get(int(layer));
                        
                        Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                        
                        if (config.hidden_layers-1 == layer) //Last layer
                         {                     
                           LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                            
                           DELTA.Col(LossGradient, 0);
                         }
                         
                        else
                         {
                           W = this.Weights_tensor.Get(layer+1);
                           
                           DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                         }
                   
                        W = this.Weights_tensor.Get(layer);
                        B = this.Bias_tensor.Get(layer);
                        
                        dB = DELTA;
                        dW = DELTA.MatMul(temp_inputs.Transpose());  
                         
                      //--- Updating the weights using the chosen algorithm
                        
                        optimizer_weights.update(W, dW);
                        optimizer_bias.update(B, dB);
                        
                        this.Weights_tensor.Add(W, layer);
                        this.Bias_tensor.Add(B, layer);  
                     
                    }
                 }  
             }
             
      //--- End of an epoch
      
        vector validation_loss(cv_tensor.SIZE);
        vector validation_acc(cv_tensor.SIZE);
        for (ulong i=0; i<cv_tensor.SIZE; i++)
          {
            validation_data = cv_tensor.Get(i);
            MatrixExtend::XandYSplitMatrices(validation_data, validation_x, validation_y);
            
            vector val_preds = this.predict_bin(validation_x);;
            
            validation_loss[i] = val_preds.Loss(validation_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
            validation_acc[i] = Metrics::accuracy_score(validation_y, val_preds);
          }
      
//---

        pred_v = this.predict_bin(x);
        
        if (batch_size==0)
          {              
              backprop_struct.training_loss[epoch] = pred_v.Loss(y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
        else
          {
              backprop_struct.training_loss[epoch] = batch_loss.Mean();
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
          
        double epoch_stop = GetTickCount();  
        
        printf("--> Epoch [%d/%d] training -> loss %.8f accuracy_score %.3f validation -> loss %.5f accuracy_score %.3f | Elapsed %s ",epoch+1,epochs,backprop_struct.training_loss[epoch],Metrics::accuracy_score(y, pred_v),backprop_struct.validation_loss[epoch],validation_acc.Mean(),this.ConvertTime((epoch_stop-epoch_start)/1000.0)); 
    }  
  
  
  if (CheckPointer(adam)!=POINTER_INVALID)  
    delete adam;
      
  return backprop_struct;
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
backprop CPatternNets::backpropagation(const matrix &x,const vector &y,const uint epochs,OptimizerBGD *bgd,uint batch_size=0,bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {
   
//--- dataframe configurations
   
   ulong inputs = x.Cols();
   ulong rows  = x.Rows();

   config.classes_in_data = MatrixExtend::Unique(y);
   config.outputs = (int)config.classes_in_data.Size();
   config.inputs = inputs;
   
//---

   isBackProp = false;
   
   backprop backprop_struct;
   backprop_struct.Init(epochs);
   
   if (y.Size() != rows )
     {
        Print(__FUNCTION__," FATAL | Number of rows  in the x matrix is not the same the y vector size ");
        
        return backprop_struct;
     }
     
     vector v2 = {(int)config.outputs};
     config.HL_NODES = MatrixExtend::concatenate(config.HL_NODES, v2); //Appending the output layer size to the hidden layers 

     config.w_config.Resize(config.HL_NODES.Size()); 
     config.hidden_layers = config.HL_NODES.Size();

//--- Initiating the Optimizer
   
    OptimizerBGD optimizer_weights = bgd;
    OptimizerBGD optimizer_bias = bgd;

//--- Generating the weights and biases
    
     this.Weights_tensor = new CTensors((uint)config.hidden_layers);
     this.Bias_tensor = new CTensors((uint)config.hidden_layers);
     this.Input_tensor = new CTensors((uint)config.hidden_layers);
     this.Output_tensor = new CTensors((uint)config.hidden_layers);
     
     ulong layer_input = config.inputs; 
     
     matrix W, B;
     
     for (ulong i=0; i<config.hidden_layers; i++)
       {
          config.w_config[i] = layer_input*config.HL_NODES[i];
          
          W = WeightsInitializers::Initializer((ulong)config.HL_NODES[i], layer_input, WEIGHT_INIT);
          this.Weights_tensor.Add(W, i);
          
          B.Init((ulong)config.HL_NODES[i], 1);
          B.Fill(0.0); //Random weight initialization can lead to symmetry issues where all neurons in a layer learn identical features. Zeros for biases help break this symmetry, allowing each neuron to learn independently during training.
          this.Bias_tensor.Add(B, i);
          
          layer_input = (ulong)config.HL_NODES[i];
       }
   
   isBackProp = true;
    
//---      
   
    matrix DELTA = {};
    vector Forecast_v={};
    
    matrix  temp_inputs ={};
   
    
    matrix dB = {}; //Bias Derivatives
    matrix dW = {}; //Weight Derivatives 
    
    
    matrix ONEHOT = MatrixExtend::OneHotEncoding(y); //ONE HOT ENCODED MATRIX
    
//---
    
    CCrossValidation cross_validation;        
    CTensors *cv_tensor;
    matrix validation_data = MatrixExtend::concatenate(x, y);
    matrix validation_x;
    vector validation_y;
     
    cv_tensor =  cross_validation.KFoldCV(validation_data, 10);
    
    for (ulong epoch=0; epoch<epochs && !IsStopped(); epoch++) //Run an entire epoch
      { 
       
        double epoch_start = GetTickCount(); 
        
        uint num_batches = (uint)MathFloor(x.Rows()/(batch_size+DBL_EPSILON));
        
        vector batch_loss(num_batches), 
               batch_accuracy(num_batches);
               
        vector actual_v ={}, pred_v={}, LossGradient = {};
        
        if (batch_size>0)
         {  
            for (uint batch=0, batch_start=0, batch_end=batch_size; batch<num_batches; batch++, batch_start+=batch_size, batch_end=(batch_start+batch_size-1))
               {
                  
                  matrix batch_x = MatrixExtend::Get(x, batch_start, batch_end-1);
                  vector batch_y = MatrixExtend::Get(y, batch_start, batch_end-1);
                  
                  rows = batch_x.Rows();            
                  matrix ONEHOT = MatrixExtend::OneHotEncoding(batch_y); //ONE HOT ENCODED MATRIX      
                  
                    for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                      {
                        pred_v = predict_proba(batch_x.Row(iter));
                        actual_v = ONEHOT.Row(iter);
                        
            //---
                        
                         DELTA.Resize(config.outputs, 1);
                         
                         for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                           {                
                              Partial_Derivatives = this.Output_tensor.Get(int(layer));
                              temp_inputs = this.Input_tensor.Get(int(layer));
                              
                              Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                              
                              if (config.hidden_layers-1 == layer) //Last layer
                               {                     
                                 LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                                  
                                 DELTA.Col(LossGradient, 0);
                               }
                               
                              else
                               {
                                 W = this.Weights_tensor.Get(layer+1);
                                 
                                 DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                               }
                         
                           W = this.Weights_tensor.Get(layer);
                           B = this.Bias_tensor.Get(layer);
                           
                           dB = DELTA;
                           dW = DELTA.MatMul(temp_inputs.Transpose());  
                            
                         //--- Updating the weights using the chosen algorithm
                           
                           optimizer_weights.update(W, dW);
                           optimizer_bias.update(B, dB);
                           
                           this.Weights_tensor.Add(W, layer);
                           this.Bias_tensor.Add(B, layer);  
                        }
                   }
                 
                 pred_v = predict_bin(batch_x);
                 
                 batch_loss[batch] = pred_v.Loss(batch_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
                 batch_loss[batch] = MathIsValidNumber(batch_loss[batch]) ? (batch_loss[batch]>1e6 ? 1e6 : batch_loss[batch]) : 1e6; //Check for nan and return some large value if it is nan
                 
                 batch_accuracy[batch] = Metrics::accuracy_score(batch_y,pred_v);
                 
                 if (show_batch_progress)
                  printf("----> batch[%d/%d] batch-loss %.5f accuracy %.3f",batch+1,num_batches,batch_loss[batch],batch_accuracy[batch]);  
              }
           }    
         else
           {
              for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                {
                  pred_v = predict_proba(x.Row(iter));
                  actual_v = ONEHOT.Row(iter);
                  
      //---
                   DELTA.Resize(config.outputs, 1);
                   
                   for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                     {                
                        Partial_Derivatives = this.Output_tensor.Get(int(layer));
                        temp_inputs = this.Input_tensor.Get(int(layer));
                        
                        Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                        
                        if (config.hidden_layers-1 == layer) //Last layer
                         {                     
                           LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                            
                           DELTA.Col(LossGradient, 0);
                         }
                         
                        else
                         {
                           W = this.Weights_tensor.Get(layer+1);
                           
                           DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                         }
                   
                        W = this.Weights_tensor.Get(layer);
                        B = this.Bias_tensor.Get(layer);
                        
                        dB = DELTA;
                        dW = DELTA.MatMul(temp_inputs.Transpose());  
                         
                      //--- Updating the weights using the chosen algorithm
                        
                        optimizer_weights.update(W, dW);
                        optimizer_bias.update(B, dB);
                        
                        this.Weights_tensor.Add(W, layer);
                        this.Bias_tensor.Add(B, layer);  
                     
                    }
                 }  
             }
             
      //--- End of an epoch
      
        vector validation_loss(cv_tensor.SIZE);
        vector validation_acc(cv_tensor.SIZE);
        for (ulong i=0; i<cv_tensor.SIZE; i++)
          {
            validation_data = cv_tensor.Get(i);
            MatrixExtend::XandYSplitMatrices(validation_data, validation_x, validation_y);
            
            vector val_preds = this.predict_bin(validation_x);;
            
            validation_loss[i] = val_preds.Loss(validation_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
            validation_acc[i] = Metrics::accuracy_score(validation_y, val_preds);
          }
      
//---

        pred_v = this.predict_bin(x);
        
        if (batch_size==0)
          {              
              backprop_struct.training_loss[epoch] = pred_v.Loss(y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
        else
          {
              backprop_struct.training_loss[epoch] = batch_loss.Mean();
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
          
        double epoch_stop = GetTickCount();  
        
        printf("--> Epoch [%d/%d] training -> loss %.8f accuracy_score %.3f validation -> loss %.5f accuracy_score %.3f | Elapsed %s ",epoch+1,epochs,backprop_struct.training_loss[epoch],Metrics::accuracy_score(y, pred_v),backprop_struct.validation_loss[epoch],validation_acc.Mean(),this.ConvertTime((epoch_stop-epoch_start)/1000.0)); 
    }  
  
  
  if (CheckPointer(bgd)!=POINTER_INVALID)  
    delete bgd;
      
  return backprop_struct;
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
backprop CPatternNets::backpropagation(const matrix &x,const vector &y,const uint epochs,OptimizerNadam* nadam,uint batch_size=0,bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {
   
//--- dataframe configurations
   
   ulong inputs = x.Cols();
   ulong rows  = x.Rows();

   config.classes_in_data = MatrixExtend::Unique(y);
   config.outputs = (int)config.classes_in_data.Size();
   config.inputs = inputs;
   
//---

   isBackProp = false;
   
   backprop backprop_struct;
   backprop_struct.Init(epochs);
   
   if (y.Size() != rows )
     {
        Print(__FUNCTION__," FATAL | Number of rows  in the x matrix is not the same the y vector size ");
        
        return backprop_struct;
     }
     
     vector v2 = {(int)config.outputs};
     config.HL_NODES = MatrixExtend::concatenate(config.HL_NODES, v2); //Appending the output layer size to the hidden layers 

     config.w_config.Resize(config.HL_NODES.Size()); 
     config.hidden_layers = config.HL_NODES.Size();

//--- Initiating the Optimizer
   
    OptimizerNadam optimizer_weights = nadam;
    OptimizerNadam optimizer_bias = nadam;

//--- Generating the weights and biases
    
     this.Weights_tensor = new CTensors((uint)config.hidden_layers);
     this.Bias_tensor = new CTensors((uint)config.hidden_layers);
     this.Input_tensor = new CTensors((uint)config.hidden_layers);
     this.Output_tensor = new CTensors((uint)config.hidden_layers);
     
     ulong layer_input = config.inputs; 
     
     matrix W, B;
     
     for (ulong i=0; i<config.hidden_layers; i++)
       {
          config.w_config[i] = layer_input*config.HL_NODES[i];
          
          W = WeightsInitializers::Initializer((ulong)config.HL_NODES[i], layer_input, WEIGHT_INIT);
          this.Weights_tensor.Add(W, i);
          
          B.Init((ulong)config.HL_NODES[i], 1);
          B.Fill(0.0); //Random weight initialization can lead to symmetry issues where all neurons in a layer learn identical features. Zeros for biases help break this symmetry, allowing each neuron to learn independently during training.
          this.Bias_tensor.Add(B, i);
          
          layer_input = (ulong)config.HL_NODES[i];
       }
   
   isBackProp = true;
    
//---      
   
    matrix DELTA = {};
    vector Forecast_v={};
    
    matrix  temp_inputs ={};
   
    
    matrix dB = {}; //Bias Derivatives
    matrix dW = {}; //Weight Derivatives 
    
    
    matrix ONEHOT = MatrixExtend::OneHotEncoding(y); //ONE HOT ENCODED MATRIX
    
//---
    
    CCrossValidation cross_validation;        
    CTensors *cv_tensor;
    matrix validation_data = MatrixExtend::concatenate(x, y);
    matrix validation_x;
    vector validation_y;
     
    cv_tensor =  cross_validation.KFoldCV(validation_data, 10);
    
    for (ulong epoch=0; epoch<epochs && !IsStopped(); epoch++) //Run an entire epoch
      { 
       
        double epoch_start = GetTickCount(); 
        
        uint num_batches = (uint)MathFloor(x.Rows()/(batch_size+DBL_EPSILON));
        
        vector batch_loss(num_batches), 
               batch_accuracy(num_batches);
               
        vector actual_v ={}, pred_v={}, LossGradient = {};
        
        if (batch_size>0)
         {  
            for (uint batch=0, batch_start=0, batch_end=batch_size; batch<num_batches; batch++, batch_start+=batch_size, batch_end=(batch_start+batch_size-1))
               {
                  
                  matrix batch_x = MatrixExtend::Get(x, batch_start, batch_end-1);
                  vector batch_y = MatrixExtend::Get(y, batch_start, batch_end-1);
                  
                  rows = batch_x.Rows();            
                  matrix ONEHOT = MatrixExtend::OneHotEncoding(batch_y); //ONE HOT ENCODED MATRIX      
                  
                    for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                      {
                        pred_v = predict_proba(batch_x.Row(iter));
                        actual_v = ONEHOT.Row(iter);
                        
            //---
                        
                         DELTA.Resize(config.outputs, 1);
                         
                         for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                           {                
                              Partial_Derivatives = this.Output_tensor.Get(int(layer));
                              temp_inputs = this.Input_tensor.Get(int(layer));
                              
                              Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                              
                              if (config.hidden_layers-1 == layer) //Last layer
                               {                     
                                 LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                                  
                                 DELTA.Col(LossGradient, 0);
                               }
                               
                              else
                               {
                                 W = this.Weights_tensor.Get(layer+1);
                                 
                                 DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                               }
                         
                           W = this.Weights_tensor.Get(layer);
                           B = this.Bias_tensor.Get(layer);
                           
                           dB = DELTA;
                           dW = DELTA.MatMul(temp_inputs.Transpose());  
                            
                         //--- Updating the weights using the chosen algorithm
                           
                           optimizer_weights.update(W, dW);
                           optimizer_bias.update(B, dB);
                           
                           this.Weights_tensor.Add(W, layer);
                           this.Bias_tensor.Add(B, layer);  
                        }
                   }
                 
                 pred_v = predict_bin(batch_x);
                 
                 batch_loss[batch] = pred_v.Loss(batch_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
                 batch_loss[batch] = MathIsValidNumber(batch_loss[batch]) ? (batch_loss[batch]>1e6 ? 1e6 : batch_loss[batch]) : 1e6; //Check for nan and return some large value if it is nan
                 
                 batch_accuracy[batch] = Metrics::accuracy_score(batch_y,pred_v);
                 
                 if (show_batch_progress)
                  printf("----> batch[%d/%d] batch-loss %.5f accuracy %.3f",batch+1,num_batches,batch_loss[batch],batch_accuracy[batch]);  
              }
           }    
         else
           {
              for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                {
                  pred_v = predict_proba(x.Row(iter));
                  actual_v = ONEHOT.Row(iter);
                  
      //---
                   DELTA.Resize(config.outputs, 1);
                   
                   for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                     {                
                        Partial_Derivatives = this.Output_tensor.Get(int(layer));
                        temp_inputs = this.Input_tensor.Get(int(layer));
                        
                        Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                        
                        if (config.hidden_layers-1 == layer) //Last layer
                         {                     
                           LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                            
                           DELTA.Col(LossGradient, 0);
                         }
                         
                        else
                         {
                           W = this.Weights_tensor.Get(layer+1);
                           
                           DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                         }
                   
                        W = this.Weights_tensor.Get(layer);
                        B = this.Bias_tensor.Get(layer);
                        
                        dB = DELTA;
                        dW = DELTA.MatMul(temp_inputs.Transpose());  
                         
                      //--- Updating the weights using the chosen algorithm
                        
                        optimizer_weights.update(W, dW);
                        optimizer_bias.update(B, dB);
                        
                        this.Weights_tensor.Add(W, layer);
                        this.Bias_tensor.Add(B, layer);  
                     
                    }
                 }  
             }
             
      //--- End of an epoch
      
        vector validation_loss(cv_tensor.SIZE);
        vector validation_acc(cv_tensor.SIZE);
        for (ulong i=0; i<cv_tensor.SIZE; i++)
          {
            validation_data = cv_tensor.Get(i);
            MatrixExtend::XandYSplitMatrices(validation_data, validation_x, validation_y);
            
            vector val_preds = this.predict_bin(validation_x);;
            
            validation_loss[i] = val_preds.Loss(validation_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
            validation_acc[i] = Metrics::accuracy_score(validation_y, val_preds);
          }
      
//---

        pred_v = this.predict_bin(x);
        
        if (batch_size==0)
          {              
              backprop_struct.training_loss[epoch] = pred_v.Loss(y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
        else
          {
              backprop_struct.training_loss[epoch] = batch_loss.Mean();
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
          
        double epoch_stop = GetTickCount();  
        
        printf("--> Epoch [%d/%d] training -> loss %.8f accuracy_score %.3f validation -> loss %.5f accuracy_score %.3f | Elapsed %s ",epoch+1,epochs,backprop_struct.training_loss[epoch],Metrics::accuracy_score(y, pred_v),backprop_struct.validation_loss[epoch],validation_acc.Mean(),this.ConvertTime((epoch_stop-epoch_start)/1000.0));    
    }  
  
  
  if (CheckPointer(nadam)!=POINTER_INVALID)  
    delete nadam;
      
  return backprop_struct;
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
backprop CPatternNets::backpropagation(const matrix &x,const vector &y,const uint epochs,OptimizerRMSprop *rmsprop,uint batch_size=0,bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {
   
//--- dataframe configurations
   
   ulong inputs = x.Cols();
   ulong rows  = x.Rows();

   config.classes_in_data = MatrixExtend::Unique(y);
   config.outputs = (int)config.classes_in_data.Size();
   config.inputs = inputs;
   
//---

   isBackProp = false;
   
   backprop backprop_struct;
   backprop_struct.Init(epochs);
   
   if (y.Size() != rows )
     {
        Print(__FUNCTION__," FATAL | Number of rows  in the x matrix is not the same the y vector size ");
        
        return backprop_struct;
     }
     
     vector v2 = {(int)config.outputs};
     config.HL_NODES = MatrixExtend::concatenate(config.HL_NODES, v2); //Appending the output layer size to the hidden layers 

     config.w_config.Resize(config.HL_NODES.Size()); 
     config.hidden_layers = config.HL_NODES.Size();

//--- Initiating the Optimizer
   
    OptimizerRMSprop optimizer_weights = rmsprop;
    OptimizerRMSprop optimizer_bias = rmsprop;

//--- Generating the weights and biases
    
     this.Weights_tensor = new CTensors((uint)config.hidden_layers);
     this.Bias_tensor = new CTensors((uint)config.hidden_layers);
     this.Input_tensor = new CTensors((uint)config.hidden_layers);
     this.Output_tensor = new CTensors((uint)config.hidden_layers);
     
     ulong layer_input = config.inputs; 
     
     matrix W, B;
     
     for (ulong i=0; i<config.hidden_layers; i++)
       {
          config.w_config[i] = layer_input*config.HL_NODES[i];
          
          W = WeightsInitializers::Initializer((ulong)config.HL_NODES[i], layer_input, WEIGHT_INIT);
          this.Weights_tensor.Add(W, i);
          
          B.Init((ulong)config.HL_NODES[i], 1);
          B.Fill(0.0); //Random weight initialization can lead to symmetry issues where all neurons in a layer learn identical features. Zeros for biases help break this symmetry, allowing each neuron to learn independently during training.
          this.Bias_tensor.Add(B, i);
          
          layer_input = (ulong)config.HL_NODES[i];
       }
   
   isBackProp = true;
    
//---      
   
    matrix DELTA = {};
    vector Forecast_v={};
    
    matrix  temp_inputs ={};
   
    
    matrix dB = {}; //Bias Derivatives
    matrix dW = {}; //Weight Derivatives 
    
    
    matrix ONEHOT = MatrixExtend::OneHotEncoding(y); //ONE HOT ENCODED MATRIX
    
//---
    
    CCrossValidation cross_validation;        
    CTensors *cv_tensor;
    matrix validation_data = MatrixExtend::concatenate(x, y);
    matrix validation_x;
    vector validation_y;
     
    cv_tensor =  cross_validation.KFoldCV(validation_data, 10);
    
    for (ulong epoch=0; epoch<epochs && !IsStopped(); epoch++) //Run an entire epoch
      { 
       
        double epoch_start = GetTickCount(); 
        
        uint num_batches = (uint)MathFloor(x.Rows()/(batch_size+DBL_EPSILON));
        
        vector batch_loss(num_batches), 
               batch_accuracy(num_batches);
               
        vector actual_v ={}, pred_v={}, LossGradient = {};
        
        if (batch_size>0)
         {  
            for (uint batch=0, batch_start=0, batch_end=batch_size; batch<num_batches; batch++, batch_start+=batch_size, batch_end=(batch_start+batch_size-1))
               {
                  
                  matrix batch_x = MatrixExtend::Get(x, batch_start, batch_end-1);
                  vector batch_y = MatrixExtend::Get(y, batch_start, batch_end-1);
                  
                  rows = batch_x.Rows();            
                  matrix ONEHOT = MatrixExtend::OneHotEncoding(batch_y); //ONE HOT ENCODED MATRIX      
                  
                    for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                      {
                        pred_v = predict_proba(batch_x.Row(iter));
                        actual_v = ONEHOT.Row(iter);
                        
            //---
                        
                         DELTA.Resize(config.outputs, 1);
                         
                         for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                           {                
                              Partial_Derivatives = this.Output_tensor.Get(int(layer));
                              temp_inputs = this.Input_tensor.Get(int(layer));
                              
                              Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                              
                              if (config.hidden_layers-1 == layer) //Last layer
                               {                     
                                 LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                                  
                                 DELTA.Col(LossGradient, 0);
                               }
                               
                              else
                               {
                                 W = this.Weights_tensor.Get(layer+1);
                                 
                                 DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                               }
                         
                           W = this.Weights_tensor.Get(layer);
                           B = this.Bias_tensor.Get(layer);
                           
                           dB = DELTA;
                           dW = DELTA.MatMul(temp_inputs.Transpose());  
                            
                         //--- Updating the weights using the chosen algorithm
                           
                           optimizer_weights.update(W, dW);
                           optimizer_bias.update(B, dB);
                           
                           this.Weights_tensor.Add(W, layer);
                           this.Bias_tensor.Add(B, layer);  
                        }
                   }
                 
                 pred_v = predict_bin(batch_x);
                 
                 batch_loss[batch] = pred_v.Loss(batch_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
                 batch_loss[batch] = MathIsValidNumber(batch_loss[batch]) ? (batch_loss[batch]>1e6 ? 1e6 : batch_loss[batch]) : 1e6; //Check for nan and return some large value if it is nan
                 
                 batch_accuracy[batch] = Metrics::accuracy_score(batch_y,pred_v);
                 
                 if (show_batch_progress)
                  printf("----> batch[%d/%d] batch-loss %.5f accuracy %.3f",batch+1,num_batches,batch_loss[batch],batch_accuracy[batch]);  
              }
           }    
         else
           {
              for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                {
                  pred_v = predict_proba(x.Row(iter));
                  actual_v = ONEHOT.Row(iter);
                  
      //---
                   DELTA.Resize(config.outputs, 1);
                   
                   for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                     {                
                        Partial_Derivatives = this.Output_tensor.Get(int(layer));
                        temp_inputs = this.Input_tensor.Get(int(layer));
                        
                        Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                        
                        if (config.hidden_layers-1 == layer) //Last layer
                         {                     
                           LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                            
                           DELTA.Col(LossGradient, 0);
                         }
                         
                        else
                         {
                           W = this.Weights_tensor.Get(layer+1);
                           
                           DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                         }
                   
                        W = this.Weights_tensor.Get(layer);
                        B = this.Bias_tensor.Get(layer);
                        
                        dB = DELTA;
                        dW = DELTA.MatMul(temp_inputs.Transpose());  
                         
                      //--- Updating the weights using the chosen algorithm
                        
                        optimizer_weights.update(W, dW);
                        optimizer_bias.update(B, dB);
                        
                        this.Weights_tensor.Add(W, layer);
                        this.Bias_tensor.Add(B, layer);  
                     
                    }
                 }  
             }
             
      //--- End of an epoch
      
        vector validation_loss(cv_tensor.SIZE);
        vector validation_acc(cv_tensor.SIZE);
        for (ulong i=0; i<cv_tensor.SIZE; i++)
          {
            validation_data = cv_tensor.Get(i);
            MatrixExtend::XandYSplitMatrices(validation_data, validation_x, validation_y);
            
            vector val_preds = this.predict_bin(validation_x);;
            
            validation_loss[i] = val_preds.Loss(validation_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
            validation_acc[i] = Metrics::accuracy_score(validation_y, val_preds);
          }
      
//---

        pred_v = this.predict_bin(x);
        
        if (batch_size==0)
          {              
              backprop_struct.training_loss[epoch] = pred_v.Loss(y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
        else
          {
              backprop_struct.training_loss[epoch] = batch_loss.Mean();
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
          
        double epoch_stop = GetTickCount();  
        
        printf("--> Epoch [%d/%d] training -> loss %.8f accuracy_score %.3f validation -> loss %.5f accuracy_score %.3f | Elapsed %s ",epoch+1,epochs,backprop_struct.training_loss[epoch],Metrics::accuracy_score(y, pred_v),backprop_struct.validation_loss[epoch],validation_acc.Mean(),this.ConvertTime((epoch_stop-epoch_start)/1000.0)); 
    }  
  
  
  if (CheckPointer(rmsprop)!=POINTER_INVALID)  
    delete rmsprop;
      
  return backprop_struct;
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
backprop CPatternNets::backpropagation(const matrix &x,const vector &y,const uint epochs,OptimizerSGD *sgd,uint batch_size=0,bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {
   
//--- dataframe configurations
   
   ulong inputs = x.Cols();
   ulong rows  = x.Rows();

   config.classes_in_data = MatrixExtend::Unique(y);
   config.outputs = (int)config.classes_in_data.Size();
   config.inputs = inputs;
   
//---

   isBackProp = false;
   
   backprop backprop_struct;
   backprop_struct.Init(epochs);
   
   if (y.Size() != rows )
     {
        Print(__FUNCTION__," FATAL | Number of rows  in the x matrix is not the same the y vector size ");
        
        return backprop_struct;
     }
     
     vector v2 = {(int)config.outputs};
     config.HL_NODES = MatrixExtend::concatenate(config.HL_NODES, v2); //Appending the output layer size to the hidden layers 

     config.w_config.Resize(config.HL_NODES.Size()); 
     config.hidden_layers = config.HL_NODES.Size();

//--- Initiating the Optimizer
   
    OptimizerSGD optimizer_weights = sgd;
    OptimizerSGD optimizer_bias = sgd;

//--- Generating the weights and biases
    
     this.Weights_tensor = new CTensors((uint)config.hidden_layers);
     this.Bias_tensor = new CTensors((uint)config.hidden_layers);
     this.Input_tensor = new CTensors((uint)config.hidden_layers);
     this.Output_tensor = new CTensors((uint)config.hidden_layers);
     
     ulong layer_input = config.inputs; 
     
     matrix W, B;
     
     for (ulong i=0; i<config.hidden_layers; i++)
       {
          config.w_config[i] = layer_input*config.HL_NODES[i];
          
          W = WeightsInitializers::Initializer((ulong)config.HL_NODES[i], layer_input, WEIGHT_INIT);
          this.Weights_tensor.Add(W, i);
          
          B.Init((ulong)config.HL_NODES[i], 1);
          B.Fill(0.0); //Random weight initialization can lead to symmetry issues where all neurons in a layer learn identical features. Zeros for biases help break this symmetry, allowing each neuron to learn independently during training.
          this.Bias_tensor.Add(B, i);
          
          layer_input = (ulong)config.HL_NODES[i];
       }
   
   isBackProp = true;
    
//---      
   
    matrix DELTA = {};
    vector Forecast_v={};
    
    matrix  temp_inputs ={};
   
    
    matrix dB = {}; //Bias Derivatives
    matrix dW = {}; //Weight Derivatives 
    
    
    matrix ONEHOT = MatrixExtend::OneHotEncoding(y); //ONE HOT ENCODED MATRIX
    
//---
    
    CCrossValidation cross_validation;        
    CTensors *cv_tensor;
    matrix validation_data = MatrixExtend::concatenate(x, y);
    matrix validation_x;
    vector validation_y;
     
    cv_tensor =  cross_validation.KFoldCV(validation_data, 10);
    
    for (ulong epoch=0; epoch<epochs && !IsStopped(); epoch++) //Run an entire epoch
      { 
       
        double epoch_start = GetTickCount(); 
        
        uint num_batches = (uint)MathFloor(x.Rows()/(batch_size+DBL_EPSILON));
        
        vector batch_loss(num_batches), 
               batch_accuracy(num_batches);
               
        vector actual_v ={}, pred_v={}, LossGradient = {};
        
        if (batch_size>0)
         {  
            for (uint batch=0, batch_start=0, batch_end=batch_size; batch<num_batches; batch++, batch_start+=batch_size, batch_end=(batch_start+batch_size-1))
               {
                  
                  matrix batch_x = MatrixExtend::Get(x, batch_start, batch_end-1);
                  vector batch_y = MatrixExtend::Get(y, batch_start, batch_end-1);
                  
                  rows = batch_x.Rows();            
                  matrix ONEHOT = MatrixExtend::OneHotEncoding(batch_y); //ONE HOT ENCODED MATRIX      
                  
                    for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                      {
                        pred_v = predict_proba(batch_x.Row(iter));
                        actual_v = ONEHOT.Row(iter);
                        
            //---
                        
                         DELTA.Resize(config.outputs, 1);
                         
                         for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                           {                
                              Partial_Derivatives = this.Output_tensor.Get(int(layer));
                              temp_inputs = this.Input_tensor.Get(int(layer));
                              
                              Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                              
                              if (config.hidden_layers-1 == layer) //Last layer
                               {                     
                                 LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                                  
                                 DELTA.Col(LossGradient, 0);
                               }
                               
                              else
                               {
                                 W = this.Weights_tensor.Get(layer+1);
                                 
                                 DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                               }
                         
                           W = this.Weights_tensor.Get(layer);
                           B = this.Bias_tensor.Get(layer);
                           
                           dB = DELTA;
                           dW = DELTA.MatMul(temp_inputs.Transpose());  
                            
                         //--- Updating the weights using the chosen algorithm
                           
                           optimizer_weights.update(W, dW);
                           optimizer_bias.update(B, dB);
                           
                           this.Weights_tensor.Add(W, layer);
                           this.Bias_tensor.Add(B, layer);  
                        }
                   }
                 
                 pred_v = predict_bin(batch_x);
                 
                 batch_loss[batch] = pred_v.Loss(batch_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
                 batch_loss[batch] = MathIsValidNumber(batch_loss[batch]) ? (batch_loss[batch]>1e6 ? 1e6 : batch_loss[batch]) : 1e6; //Check for nan and return some large value if it is nan
                 
                 batch_accuracy[batch] = Metrics::accuracy_score(batch_y,pred_v);
                 
                 if (show_batch_progress)
                  printf("----> batch[%d/%d] batch-loss %.5f accuracy %.3f",batch+1,num_batches,batch_loss[batch],batch_accuracy[batch]);  
              }
           }    
         else
           {
              for (ulong iter=0; iter<rows ; iter++) //replace to rows 
                {
                  pred_v = predict_proba(x.Row(iter));
                  actual_v = ONEHOT.Row(iter);
                  
      //---
                   DELTA.Resize(config.outputs, 1);
                   
                   for (int layer=(int)config.hidden_layers-1; layer>=0 && !IsStopped(); layer--)
                     {                
                        Partial_Derivatives = this.Output_tensor.Get(int(layer));
                        temp_inputs = this.Input_tensor.Get(int(layer));
                        
                        Partial_Derivatives.Derivative(Partial_Derivatives, (config.hidden_layers-1 == layer)? AF_SOFTMAX : ENUM_ACTIVATION_FUNCTION(this.config.activation_fx));
                        
                        if (config.hidden_layers-1 == layer) //Last layer
                         {                     
                           LossGradient = pred_v.LossGradient(actual_v, ENUM_LOSS_FUNCTION(this.config.loss_fx));                  
                            
                           DELTA.Col(LossGradient, 0);
                         }
                         
                        else
                         {
                           W = this.Weights_tensor.Get(layer+1);
                           
                           DELTA =  (W.Transpose().MatMul(DELTA)) * Partial_Derivatives; //element-wise multiplication
                         }
                   
                        W = this.Weights_tensor.Get(layer);
                        B = this.Bias_tensor.Get(layer);
                        
                        dB = DELTA;
                        dW = DELTA.MatMul(temp_inputs.Transpose());  
                         
                      //--- Updating the weights using the chosen algorithm
                        
                        optimizer_weights.update(W, dW);
                        optimizer_bias.update(B, dB);
                        
                        this.Weights_tensor.Add(W, layer);
                        this.Bias_tensor.Add(B, layer);  
                     
                    }
                 }  
             }
             
      //--- End of an epoch
      
        vector validation_loss(cv_tensor.SIZE);
        vector validation_acc(cv_tensor.SIZE);
        for (ulong i=0; i<cv_tensor.SIZE; i++)
          {
            validation_data = cv_tensor.Get(i);
            MatrixExtend::XandYSplitMatrices(validation_data, validation_x, validation_y);
            
            vector val_preds = this.predict_bin(validation_x);;
            
            validation_loss[i] = val_preds.Loss(validation_y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
            validation_acc[i] = Metrics::accuracy_score(validation_y, val_preds);
          }
      
//---
              
        pred_v = this.predict_bin(x);
        
        if (batch_size==0)
          {
              
              backprop_struct.training_loss[epoch] = pred_v.Loss(y, ENUM_LOSS_FUNCTION(this.config.loss_fx));
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
        else
          {
              backprop_struct.training_loss[epoch] = batch_loss.Mean();
              backprop_struct.training_loss[epoch] = MathIsValidNumber(backprop_struct.training_loss[epoch]) ? (backprop_struct.training_loss[epoch]>1e6 ? 1e6 : backprop_struct.training_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
              backprop_struct.validation_loss[epoch] = validation_loss.Mean();
              backprop_struct.validation_loss[epoch] = MathIsValidNumber(backprop_struct.validation_loss[epoch]) ? (backprop_struct.validation_loss[epoch]>1e6 ? 1e6 : backprop_struct.validation_loss[epoch]) : 1e6; //Check for nan and return some large value if it is nan
          }
          
        double epoch_stop = GetTickCount();  
        
        printf("--> Epoch [%d/%d] training -> loss %.8f accuracy_score %.3f validation -> loss %.5f accuracy_score %.3f | Elapsed %s ",epoch+1,epochs,backprop_struct.training_loss[epoch],Metrics::accuracy_score(y, pred_v),backprop_struct.validation_loss[epoch],validation_acc.Mean(),this.ConvertTime((epoch_stop-epoch_start)/1000.0));  
    }  
  
  
  if (CheckPointer(sgd)!=POINTER_INVALID)  
    delete sgd;
      
  return backprop_struct;
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+

void CPatternNets::fit(const matrix &x, const vector &y, const uint epochs, OptimizerAdam *adam, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {
  if (loadedModel) 
   {
      Print("The Pretrained Model was Loaded, There is no need to train the model");
      return;
   }
  
  trained = true; //The fit method has been called
  
  vector epochs_vector(epochs);  for (uint i=0; i<epochs; i++) epochs_vector[i] = i+1;
  
  backprop backprop_struct;
  
  backprop_struct = this.backpropagation(x, y, epochs, adam, batch_size, show_batch_progress, WEIGHT_INIT); //Run backpropagation
  

  CPlots plt;
  plt.Plot("Loss vs Epochs",epochs_vector,backprop_struct.training_loss,"epochs","loss","training-loss",CURVE_LINES);
  plt.AddPlot(backprop_struct.validation_loss,"validation-loss",clrRed);
  
   while (MessageBox("Close or Cancel Loss Vs Epoch plot to proceed","Training progress",MB_OK)<0)
    Sleep(1);

  isBackProp = false;
 } 
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CPatternNets::fit(const matrix &x, const vector &y, const uint epochs, OptimizerAdaDelta *adadelta, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {

  if (loadedModel) 
   {
      Print("The Pretrained Model was Loaded, There is no need to train the model");
      return;
   }
  
  trained = true; //The fit method has been called
  
  vector epochs_vector(epochs);  for (uint i=0; i<epochs; i++) epochs_vector[i] = i+1;
  
  backprop backprop_struct;
  
  backprop_struct = this.backpropagation(x, y, epochs, adadelta, batch_size, show_batch_progress, WEIGHT_INIT); //Run backpropagation
  

  CPlots plt;
  plt.Plot("Loss vs Epochs",epochs_vector,backprop_struct.training_loss,"epochs","loss","training-loss",CURVE_LINES);
  plt.AddPlot(backprop_struct.validation_loss,"validation-loss",clrRed);
  
   while (MessageBox("Close or Cancel Loss Vs Epoch plot to proceed","Training progress",MB_OK)<0)
    Sleep(1);

  isBackProp = false;
 } 
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CPatternNets::fit(const matrix &x, const vector &y, const uint epochs, OptimizerAdaGrad *adagrad, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {

  if (loadedModel) 
   {
      Print("The Pretrained Model was Loaded, There is no need to train the model");
      return;
   }
  
  trained = true; //The fit method has been called
  
  vector epochs_vector(epochs);  for (uint i=0; i<epochs; i++) epochs_vector[i] = i+1;
  
  backprop backprop_struct;
  
  backprop_struct = this.backpropagation(x, y, epochs, adagrad, batch_size, show_batch_progress, WEIGHT_INIT); //Run backpropagation
  

  CPlots plt;
  plt.Plot("Loss vs Epochs",epochs_vector,backprop_struct.training_loss,"epochs","loss","training-loss",CURVE_LINES);
  plt.AddPlot(backprop_struct.validation_loss,"validation-loss",clrRed);
  
   while (MessageBox("Close or Cancel Loss Vs Epoch plot to proceed","Training progress",MB_OK)<0)
    Sleep(1);

  isBackProp = false;
 } 
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CPatternNets::fit(const matrix &x, const vector &y, const uint epochs, OptimizerNadam *nadam, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {

  if (loadedModel) 
   {
      Print("The Pretrained Model was Loaded, There is no need to train the model");
      return;
   }
  
  trained = true; //The fit method has been called
  
  vector epochs_vector(epochs);  for (uint i=0; i<epochs; i++) epochs_vector[i] = i+1;
  
  backprop backprop_struct;
  
  backprop_struct = this.backpropagation(x, y, epochs, nadam, batch_size, show_batch_progress, WEIGHT_INIT); //Run backpropagation
  

  CPlots plt;
  plt.Plot("Loss vs Epochs",epochs_vector,backprop_struct.training_loss,"epochs","loss","training-loss",CURVE_LINES);
  plt.AddPlot(backprop_struct.validation_loss,"validation-loss",clrRed);
  
   while (MessageBox("Close or Cancel Loss Vs Epoch plot to proceed","Training progress",MB_OK)<0)
    Sleep(1);

  isBackProp = false;
 } 
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CPatternNets::fit(const matrix &x, const vector &y, const uint epochs, OptimizerRMSprop *rmsprop, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {

  if (loadedModel) 
   {
      Print("The Pretrained Model was Loaded, There is no need to train the model");
      return;
   }
  
  trained = true; //The fit method has been called
  
  vector epochs_vector(epochs);  for (uint i=0; i<epochs; i++) epochs_vector[i] = i+1;
  
  backprop backprop_struct;
  
  backprop_struct = this.backpropagation(x, y, epochs, rmsprop, batch_size, show_batch_progress, WEIGHT_INIT); //Run backpropagation
  

  CPlots plt;
  plt.Plot("Loss vs Epochs",epochs_vector,backprop_struct.training_loss,"epochs","loss","training-loss",CURVE_LINES);
  plt.AddPlot(backprop_struct.validation_loss,"validation-loss",clrRed);
  
   while (MessageBox("Close or Cancel Loss Vs Epoch plot to proceed","Training progress",MB_OK)<0)
    Sleep(1);

  isBackProp = false;
 } 
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CPatternNets::fit(const matrix &x, const vector &y, const uint epochs, OptimizerBGD *bgd, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {

  if (loadedModel) 
   {
      Print("The Pretrained Model was Loaded, There is no need to train the model");
      return;
   }
  
  trained = true; //The fit method has been called
  
  vector epochs_vector(epochs);  for (uint i=0; i<epochs; i++) epochs_vector[i] = i+1;
  
  backprop backprop_struct;
  
  backprop_struct = this.backpropagation(x, y, epochs, bgd, batch_size, show_batch_progress, WEIGHT_INIT); //Run backpropagation
  

  CPlots plt;
  plt.Plot("Loss vs Epochs",epochs_vector,backprop_struct.training_loss,"epochs","loss","training-loss",CURVE_LINES);
  plt.AddPlot(backprop_struct.validation_loss,"validation-loss",clrRed);
  
   while (MessageBox("Close or Cancel Loss Vs Epoch plot to proceed","Training progress",MB_OK)<0)
    Sleep(1);

  isBackProp = false;
 } 
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CPatternNets::fit(const matrix &x, const vector &y, const uint epochs, OptimizerSGD *sgd, uint batch_size=0, bool show_batch_progress=false, weight_intializers WEIGHT_INIT=WEIGHT_INTIALIZER_XAVIER)
 {

  if (loadedModel) 
   {
      Print("The Pretrained Model was Loaded, There is no need to train the model");
      return;
   }
  
  trained = true; //The fit method has been called
  
  vector epochs_vector(epochs);  for (uint i=0; i<epochs; i++) epochs_vector[i] = i+1;
  
  backprop backprop_struct;
  
  backprop_struct = this.backpropagation(x, y, epochs, sgd, batch_size, show_batch_progress, WEIGHT_INIT); //Run backpropagation
  

  CPlots plt;
  plt.Plot("Loss vs Epochs",epochs_vector,backprop_struct.training_loss,"epochs","loss","training-loss",CURVE_LINES);
  plt.AddPlot(backprop_struct.validation_loss,"validation-loss",clrRed);
  
   while (MessageBox("Close or Cancel Loss Vs Epoch plot to proceed","Training progress",MB_OK)<0)
    Sleep(1);

  isBackProp = false;
 }  
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool CPatternNets::save_model(string dir)
 {
   if (!trained) //ensure the model was trained
    {
      printf("%s train this model before attempting to save it",__FUNCTION__);
      return false;
    }
   
   if (MQLInfoInteger(MQL_DEBUG))
    printf("<<<<< SAving regressor nn-model >>>>>");
   
//--- Save the architecture | Hidden layers
   
   if (!MatrixExtend::write_bin(config.HL_NODES, dir+"\\architecture.bin"))
    {
      printf("%s failed to save nn-architecture",__FUNCTION__);
      return false;
    }
   
   if (MQLInfoInteger(MQL_DEBUG))
     Print("Architecture: ",config.HL_NODES);
   
//--- Save the weights 
  
  if (MQLInfoInteger(MQL_DEBUG))
    {
      Print("Saving Weights tensor");
      Weights_tensor.Print_();
    }
 
 { 
  vector v = {};
  for (ulong i=0; i<this.Weights_tensor.SIZE; i++)
   {
      matrix w = this.Weights_tensor.Get(i);
      vector v2 = MatrixExtend::MatrixToVector(w);
      v = MatrixExtend::concatenate(v, v2); //Save all the weights to one long vector
   }
   
  if (!MatrixExtend::write_bin(v, dir+"\\weights.bin"))
    {
      printf("%s failed to save nn-weights",__FUNCTION__);
      return false;
    }
  }
    
//--- Save the bias
   
  if (MQLInfoInteger(MQL_DEBUG))
    {
      Print("Saving bias tensor");
      Bias_tensor.Print_();
    }
    
  {  
  vector  v = {};
  for (ulong i=0; i<this.Bias_tensor.SIZE; i++)
   {
      matrix B = this.Bias_tensor.Get(i);
      vector v2 = MatrixExtend::MatrixToVector(B);
      v = MatrixExtend::concatenate(v, v2);
   }
   
   if (!MatrixExtend::write_bin(v, dir+"\\bias.bin"))
    {
      printf("%s failed to save nn-bias",__FUNCTION__);
      return false;
    }
  } 
  
//--- Activation function

  { 
   vector v = {(int)config.activation_fx};
   
   if (!MatrixExtend::write_bin(v, dir+"\\activation.bin"))
    {
      printf("%s failed to save the activation function",__FUNCTION__);
      return false;
    }
  }
  
//--- INputs

   {
   vector v = {(int)config.inputs};
   
   if (!MatrixExtend::write_bin(v, dir+"\\inputs.bin"))
    {
      printf("%s failed to save the inputs",__FUNCTION__);
      return false;
    }
   } 

//--- Saving classes in data
   
   {
   if (!MatrixExtend::write_bin(this.config.classes_in_data, dir+"\\classes.bin"))
    {
      printf("%s failed to save the classes",__FUNCTION__);
      return false;
    }
    
    if (MQLInfoInteger(MQL_DEBUG))
      Print("Saving Classes in data: ",this.config.classes_in_data);
   }

   return true;
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CPatternNets::CPatternNets(const double &architecture[], const double &activation_[], const double &classes_[], const double &inputs[], const double &weights[], const double &bias[])
 {
//--- Load architecture

   config.HL_NODES = MatrixExtend::ArrayToVector(architecture);
   
   if (config.HL_NODES.Size()==0)
    {
      printf("%s failed read the saved nn-architecture",__FUNCTION__);
      return;
    }
   
   if (MQLInfoInteger(MQL_DEBUG))
    printf("<<<<< Loading regressor nn-model >>>>>");
    
   if (MQLInfoInteger(MQL_DEBUG))
     Print("HL NOdes ",config.HL_NODES);

//--- Load the inputs
  
   {
   vector v = MatrixExtend::ArrayToVector(inputs);
   if (v.Size()==0 || v.Size()>1)
     {
       printf("%s failed to read the inputs from the saved nn-model",__FUNCTION__);
       return;
     }
   
   config.inputs = (ulong)v[0];
   }
  
//--- Loading the weights
  
  { 
   config.hidden_layers = config.HL_NODES.Size();
   this.Weights_tensor = new CTensors((uint)config.hidden_layers);
   
   vector all_weights = MatrixExtend::ArrayToVector(weights);
   if (all_weights.Size()==0)
     {
       printf("%s Failed to read nn-weights from binary file",__FUNCTION__);
       return;
     }
   
   ulong layer_inputs = config.inputs;
   ulong w_size = 0;
   for (ulong i=0, start = 0; i<config.HL_NODES.Size(); i++)
     {
       w_size = (layer_inputs*(ulong)config.HL_NODES[i]);
       
       vector l_weights = MatrixExtend::Get(all_weights, start, (start+w_size)-1);
       
       matrix w_matrix = MatrixExtend::VectorToMatrix(l_weights, layer_inputs);
       Weights_tensor.Add(w_matrix, i);
       
       layer_inputs = (ulong)config.HL_NODES[i];
       
       start += w_size;
     }
   
  if (MQLInfoInteger(MQL_DEBUG))
    {
      Print("Weights tensor Loaded");
      Weights_tensor.Print_();
    }
 }
  
//--- Loading the bias
 
 {
   this.Bias_tensor= new CTensors((uint)config.hidden_layers);
   vector all_bias = MatrixExtend::ArrayToVector(bias);
   
   if (all_bias.Size()==0)
    {
       printf("%s Failed to read nn-bias from binary file",__FUNCTION__);
       return;
    }
  
//---
  
  ulong layer_inputs = config.inputs;
  ulong b_size = 0;
   for (ulong i=0, start=0; i<config.HL_NODES.Size(); i++)
     {
       b_size = (ulong)config.HL_NODES[i];
       vector l_bias = MatrixExtend::Get(all_bias, start, (start+b_size)-1);
       matrix b_matrix = MatrixExtend::VectorToMatrix(l_bias, 1);
       
       Bias_tensor.Add(b_matrix, i);
       
       layer_inputs = (ulong)config.HL_NODES[i];
       start += b_size;
     }

 
  if (MQLInfoInteger(MQL_DEBUG))
    {
      Print("Bias tensor Loaded");
      Bias_tensor.Print_();
    }
  }   
  
//--- Loading activation function 

  { 
   vector v = MatrixExtend::ArrayToVector(activation_);
   
   if (v.Size()==0 || v.Size()>1)
    {
      printf("%s failed to read the activation function",__FUNCTION__);
      return;
    }
   
   config.activation_fx = activation(int(v[0])); 
   
   if (MQLInfoInteger(MQL_DEBUG))
     Print("NN_activation_function ",EnumToString(ENUM_ACTIVATION_FUNCTION(config.activation_fx)));
  }

//--- Loading classes in data

  {
   this.config.classes_in_data = MatrixExtend::ArrayToVector(classes_);
   
   if (this.config.classes_in_data.Size()==0)
    {
      printf("%s failed to read the classes from bin file",__FUNCTION__);
      return;
    }
    
    if (MQLInfoInteger(MQL_DEBUG))
      Print("Loaded Classes in data: ",this.config.classes_in_data);
  }
//---

   this.loadedModel = true;
   this.trained = true;
 }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
